
// depmain.cpp
// Copyright 2022 Matthew Rickard
// This file is part of dep

// dep is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "logMsg.h"
#include <fstream>
#include "dep.h"
#include <string.h>
#include "MacroNames.h"
#include <sys/stat.h>
#include "AbstractProcess.h"
#include "matches.h"
#include "DepTree.h"

int atoidefault(const char *s, int def) {
  if (isdigit(*s))
    return atoi(s);
  return def;
}

const char *bs = "dep version 0.3 copyright 2022 Matthew Rickard. Buildstamp " __DATE__ " " __TIME__;

int main(int argc, const char *argv[]) {
  Cmdspec cmdspec("-f: -h-help--help -V--version -r--remake -a--remakeall -d--dryrun -q--quiet -v?"
  " --stateverbosity? --otherverbosity?"
  " --showrules -j: --treeincludes --ruleincludes --noincludes --bothincludes"
  " --maxnonincludes: --sense: --nocache --confirmchanges",
    0, Position(__FILE__, __LINE__));
  AutoCmdLine cmdline(&cmdspec);
  cmdline.optionMap["-f"] = "DepFile";
  cmdline.processArgv(FRAGMENTTYPE_COMMAND, argc, argv);
  if (cmdline.haltline) {
    if (*argv[cmdline.paramIdx] == '-' || *argv[cmdline.paramIdx] == '/') {
      //logMsg << "Checkpoint" << std::endl;
      std::cerr << "dep: Unrecognized option " << argv[cmdline.paramIdx] << std::endl
        << "Exiting" << std::endl;
      return 1;
    }
  }
  if (cmdline.optionMap.count("-h")) {
    std::cout << "Usage: dep " << cmdspec.spec << std::endl;
    return 0;
  }
  if (cmdline.optionMap.count("-V")) {
    std::cout << "dep version 0.3" << std::endl;
    return 0;
  }
  RuleSet ruleset;

  File out("stdout", 1, std::cout);
  int processCache = 1, writeCache = 1;
  ruleset.outputFile = &out;
  if (cmdline.optionMap.count("-v")) {
    ruleset.buildStateVerbosity = atoidefault(cmdline.optionMap["-v"].c_str(), 1);
    ruleset.otherVerbosity = atoidefault(cmdline.optionMap["-v"].c_str(), 1);
  }
  if (cmdline.optionMap.count("--stateverbosity")) {
    ruleset.buildStateVerbosity = atoidefault(cmdline.optionMap["--stateverbosity"].c_str(), 1);
  }
  if (cmdline.optionMap.count("--otherverbosity")) {
    ruleset.otherVerbosity = atoidefault(cmdline.optionMap["--otherverbosity"].c_str(), 1);
  }
  if (cmdline.optionMap.count("-q")) {
    ruleset.buildStateVerbosity = -1;
    ruleset.otherVerbosity = -1;
  }
  if (cmdline.optionMap.count("--confirmchanges"))
    ruleset.confirmRuleChanges = 1;
  ruleset.optionTreeIncludes = 1;
  if (cmdline.optionMap.count("-r"))
    ruleset.optionRebuildMarked = 1;
  if (cmdline.optionMap.count("-a"))
    ruleset.optionRebuildAll = 1;
  if (cmdline.optionMap.count("-d"))
    ruleset.optionDryrun = 1;
  if (cmdline.optionMap.count("--noincludes")) {
    ruleset.optionTreeIncludes = 0;
    ruleset.optionRuleIncludes = 0;
  }
  else if (cmdline.optionMap.count("--treeincludes")) {
    ruleset.optionTreeIncludes = 1;
    ruleset.optionRuleIncludes = 0;
  }
  else if (cmdline.optionMap.count("--ruleincludes")) {
    ruleset.optionTreeIncludes = 0;
    ruleset.optionRuleIncludes = 1;
  }
  else if (cmdline.optionMap.count("--bothincludes")) {
    ruleset.optionTreeIncludes = 1;
    ruleset.optionRuleIncludes = 1;
  }
  else if (cmdline.optionMap.count("--nocache")) {
    processCache = 0;
    writeCache = 0;
    ruleset.senseNewer   = 1;
    ruleset.senseModTime = 0;
    ruleset.senseSize    = 0;
    ruleset.senseRule    = 0;
  }
  else if (cmdline.optionMap.count("--sense")) {
    std::string sense = cmdline.optionMap["--sense"];
    logMsg << logValue(ruleset.senseNewer)   << std::endl;
    logMsg << logValue(ruleset.senseModTime) << std::endl;
    logMsg << logValue(ruleset.senseSize)    << std::endl;
    logMsg << logValue(ruleset.senseRule)    << std::endl;
    logMsg << logValue(sense) << std::endl;
    ruleset.senseNewer   = sense.find("newer")   != std::string::npos;
    ruleset.senseModTime = sense.find("modtime") != std::string::npos;
    ruleset.senseSize    = sense.find("size")    != std::string::npos;
    ruleset.senseRule    = sense.find("rule")    != std::string::npos;
    logMsg << logValue(ruleset.senseNewer)   << std::endl;
    logMsg << logValue(ruleset.senseModTime) << std::endl;
    logMsg << logValue(ruleset.senseSize)    << std::endl;
    logMsg << logValue(ruleset.senseRule)    << std::endl;
    lineabort();
  }
  if (cmdline.optionMap.count("-j"))
    ruleset.maxPipelines = atoidefault(cmdline.optionMap["-j"].c_str(), 1);
  if (cmdline.optionMap.count("--maxnonincludes"))
    ruleset.maxNonIncludes = atoidefault(cmdline.optionMap["--maxnonincludes"].c_str(), 1000);

  DepTree depTree;
  depTree.ruleset = &ruleset;
  std::string filename = cmdline.optionMap["-f"].c_str();
  int result = depTree.readFile(filename);
  if (result != 0) {
    std::cerr << "dep: Can't read " << cmdline.optionMap["-f"] << std::endl;
    return 1;
  }
  if (cmdline.optionMap.count("--showrules"))
    ruleset.outputTo(std::cout);

  DepTree cacheTree;
  RuleSet cacheset;
  cacheset.buildStateVerbosity = -1;
  cacheset.otherVerbosity = -1;
  //logMsg << logValue(cacheset.buildStateVerbosity) << std::endl;
  //logMsg << logValue(cacheset.otherVerbosity) << std::endl;
  std::string cachefilename("." + filename + "Cache");
  if (processCache) {
    cacheset.outputFile = &out;
    cacheTree.ruleset = &cacheset;
    if (cacheTree.readFile(cachefilename)) {
      std::cerr << "dep: Can't read cache file " << cachefilename << std::endl;
      processCache = 0;
    }
  }

  File outputFile("stdout", 1, std::cout);
  result = ruleset.processCmd(argc - cmdline.paramIdx, argv + cmdline.paramIdx, &outputFile,
    processCache? &cacheset: 0);
  if (writeCache) {
    ruleset.combineWithCache(cacheset);
    copyFile(cachefilename, cachefilename + ".bak");
    std::ofstream ofs(cachefilename, std::ios::out | std::ios::trunc);
    //logMsg << "Outputting to " << logValue(cachefilename) << std::endl;
    ruleset.outputTo(ofs);
    //logMsg << "Output to " << logValue(cachefilename) << std::endl;
  }

  if (result == -1)
    return 1;
  if (WIFEXITED(result)) {
    if (WEXITSTATUS(result)) {
      //logMsg << "Checkpoint" << std::endl;
      std::cerr << "dep: child process exited with value " << WEXITSTATUS(result) << std::endl;
    }
    if (ruleset.failed)
      return 1;
    return WEXITSTATUS(result);
  }
  if (WIFSIGNALED(result)) {
      //logMsg << "Checkpoint" << std::endl;
    std::cerr << "dep: child process terminated by signal "
      << WTERMSIG(result) << ' ' << SignalToMacroName(WTERMSIG(result)) << std::endl;
    return 13;
  }
  if (result == 65535) {
    //logMsg << "Checkpoint" << std::endl;
    std::cerr << "dep: couldn't start build command" << std::endl;
    return 133;
  }
  logMsg << "Process terminated by something other than exit or signal "
    << logValue(result) << std::endl;
  return 113;
}

